{# Import the theme's layout. #}
{% extends "!layout.html" %}

{%- block extrahead %}
<link rel="stylesheet" type="text/css" href="{{ pathto("_static/custom.css", 1) }}">
{# Call the parent block #}
{{ super() }}
{%- endblock %}
