.MathJax .mi, .MathJax .mo {
    color: inherit;
}

pre, kdb, samp, code, .rst-content tt, .rst-content code {
    word-break: inherit;
}

code, .rst-content tt, .rst-content code {
    white-space: pre-wrap;
}

#notebook .container {
    padding-left: 0;
    width: inherit;
}

.body img {
    max-width: 100%;
}
